import logging
from typing import Any, Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series
from pydantic import BaseModel
from pydantic.networks import IPv4Address  # noqa

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class Model(BaseModel):
    ip: IPv4Address


class IPv4Type(BaseType):
    weight = 6

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for value in data:
            try:
                if not isinstance(value, str):
                    logger.debug(
                        f"Value '{value}' is not string, cannot cast to {self.name}"
                    )
                    return None
                Model(ip=value)
            except ValueError:
                logger.debug(f"Cannot cast the value '{value}' to {self.name}")
                return None
        return InferredField(inferred_type=self.name)

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        raise NotImplementedError()

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        return str(value)

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return (np.dtype("O"),)



import logging
from typing import Any, Optional, Tuple

from numpy import ndarray
from pandas import Series

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class StringType(BaseType):
    weight = 9

    def validate(self, data: ndarray) -> Optional[InferredField]:
        for value in data:
            try:
                str(value)
            except Exception:  # noqa
                logger.debug(f"Cannot cast the value '{value}' to {self.name}")
                return None
        return InferredField(inferred_type=self.name)

    @property
    def prefer_column_casting(self) -> bool:
        return True

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        return column.apply(lambda x: str(x))

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        return str(value)

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return tuple()



import math
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

from numpy import ndarray
from pandas import Series

from ..schema import InferredField


class BaseType(ABC):
    weight: int = 0

    @abstractmethod
    def validate(self, data: ndarray) -> Optional[InferredField]:
        pass

    @abstractmethod
    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        pass

    @abstractmethod
    def cast_column(
        self, column: Series, inferred_field: InferredField
    ) -> Optional[Any]:
        pass

    @abstractmethod
    def ignore_cast_for_types(self) -> Tuple[Any]:
        pass

    @property
    def name(self) -> str:
        return str(self.__class__.__name__.replace("Type", ""))

    @property
    def prefer_column_casting(self) -> bool:
        return False

    @staticmethod
    def is_empty_value(value: Any) -> bool:
        if value is None or (isinstance(value, float) and math.isnan(value)):
            return True
        return False



import csv
import locale
import logging
import time
from typing import BinaryIO, Optional, TextIO, Union

import pandas as pd
from pandas import DataFrame

from .exceptions import TooManyColumnException, exception_router
from .schema import ShmessySchema
from .types_handler import TypesHandler
from .utils import (
    _fix_column_names,
    _fix_column_names_in_df,
    _fix_column_names_in_shmessy_schema,
    _get_sample_from_csv,
    _get_sampled_df,
)

logger = logging.getLogger(__name__)


class Shmessy:
    def __init__(
        self,
        sample_size: Optional[int] = 1000,
        reader_encoding: Optional[str] = "UTF-8",
        locale_formatter: Optional[str] = "en_US",
        use_random_sample: Optional[bool] = True,
        ignore_virtual_types: Optional[bool] = False,
        max_columns_num: Optional[int] = 500,
    ) -> None:
        self.__types_handler = TypesHandler(ignore_virtual_types=ignore_virtual_types)
        self.__sample_size = sample_size
        self.__reader_encoding = reader_encoding
        self.__locale_formatter = locale_formatter
        self.__use_random_sample = use_random_sample
        self.__max_columns_num = max_columns_num

        self.__inferred_schema: Optional[ShmessySchema] = None

        locale.setlocale(
            locale.LC_ALL, f"{self.__locale_formatter}.{self.__reader_encoding}"
        )

    def get_inferred_schema(self) -> ShmessySchema:
        return self.__inferred_schema

    def infer_schema(self, df: DataFrame) -> ShmessySchema:
        start_time = time.time()
        df = _get_sampled_df(
            df=df,
            sample_size=self.__sample_size,
            random_sample=self.__use_random_sample,
        )
        columns = [
            self.__types_handler.infer_field(field_name=column, data=df[column].values)
            for column in df
        ]
        infer_duration_ms = int((time.time() - start_time) * 1000)
        inferred_schema = ShmessySchema(
            columns=columns, infer_duration_ms=infer_duration_ms
        )
        self.__inferred_schema = inferred_schema
        return inferred_schema

    def fix_schema(
        self,
        df: DataFrame,
        *,
        fix_column_names: Optional[bool] = False,
        fixed_schema: Optional[ShmessySchema] = None,
        fallback_to_string: Optional[bool] = False,
        fallback_to_null: Optional[bool] = False,
    ) -> DataFrame:
        try:
            existing_columns_num = len(df.columns)
            if existing_columns_num > self.__max_columns_num:
                raise TooManyColumnException(
                    max_columns_num=self.__max_columns_num,
                    existing_columns_num=existing_columns_num,
                )

            if fixed_schema is None:
                fixed_schema = self.infer_schema(df)

            for column in fixed_schema.columns:
                df[column.field_name] = self.__types_handler.fix_field(
                    column=df[column.field_name],
                    inferred_field=column,
                    fallback_to_string=fallback_to_string,
                    fallback_to_null=fallback_to_null,
                )

            if fix_column_names:
                mapping = _fix_column_names(df)
                df = _fix_column_names_in_df(input_df=df, mapping=mapping)
                fixed_schema = _fix_column_names_in_shmessy_schema(
                    input_schema=fixed_schema, mapping=mapping
                )

            self.__inferred_schema = fixed_schema
            return df
        except Exception as e:
            exception_router(e)

    def read_csv(
        self,
        filepath_or_buffer: Union[str, TextIO, BinaryIO],
        *,
        use_sniffer: Optional[bool] = True,
        fixed_schema: Optional[ShmessySchema] = None,
        fix_column_names: Optional[bool] = False,
        fallback_to_string: Optional[bool] = False,
        fallback_to_null: Optional[bool] = False,
    ) -> DataFrame:
        try:
            dialect = None

            if use_sniffer:
                try:
                    dialect = csv.Sniffer().sniff(
                        sample=_get_sample_from_csv(
                            filepath_or_buffer=filepath_or_buffer,
                            sample_size=self.__sample_size,
                            encoding=self.__reader_encoding,
                        ),
                        delimiters="".join([",", "\t", ";", ":"]),
                    )
                except Exception as e:  # noqa
                    logger.debug(
                        f"Could not use python sniffer to infer csv schema, Using pandas default settings: {e}"
                    )

            df = pd.read_csv(
                index_col=False,
                filepath_or_buffer=filepath_or_buffer,
                dialect=dialect() if dialect else None,
                encoding=self.__reader_encoding,
            )

            existing_columns_num = len(df.columns)
            if existing_columns_num > self.__max_columns_num:
                raise TooManyColumnException(
                    max_columns_num=self.__max_columns_num,
                    existing_columns_num=existing_columns_num,
                )

            if fixed_schema is None:
                fixed_schema = self.infer_schema(df)

            self.__inferred_schema = fixed_schema
            return self.fix_schema(
                df=df,
                fixed_schema=fixed_schema,
                fix_column_names=fix_column_names,
                fallback_to_string=fallback_to_string,
                fallback_to_null=fallback_to_null,
            )

        except Exception as e:
            exception_router(e)



import re
from typing import Any, Optional


def exception_router(exception: Exception):
    error_message = str(exception)

    match = re.match(
        r"(.*)'(.*)' codec can't decode byte (.*) in position (.*):(.*)", error_message
    )
    if match is not None:
        raise WrongEncodingException(match.group(2))

    match = re.match(
        r"(.*)Expected (.*) fields in line (.*), saw (.*)\n", error_message
    )
    if match is not None:
        raise WrongNumberOfColumnException(
            num_of_expected=match.group(2),
            num_of_saw=match.group(4),
            line_number=match.group(3),
        )
    raise exception


class ShmessyException(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class WrongNumberOfColumnException(ShmessyException):
    def __init__(self, num_of_expected: int, num_of_saw: int, line_number: int):
        super().__init__(
            f"Error in line {line_number}: Expected {num_of_expected} fields, found {num_of_saw}."
        )


class FieldCastingException(ShmessyException):
    def __init__(
        self,
        type_: str,
        bad_value: str,
        line_number: int,
        column_name: str,
        pattern: Optional[Any] = None,
    ):
        pattern_str = f"[{pattern}]" if pattern else ""
        super().__init__(
            f'Error in line {line_number:,} for column "{column_name}": '
            f'Couldn\'t cast value "{bad_value}" to type {type_}{pattern_str}.'
        )


class WrongEncodingException(ShmessyException):
    def __init__(self, expected_encoding: str):
        super().__init__(
            f"The given file cannot be read using {expected_encoding} encoding."
        )


class TooManyColumnException(ShmessyException):
    def __init__(self, existing_columns_num: int, max_columns_num: int):
        super().__init__(
            f"The input table contains {existing_columns_num} columns. "
            f"The maximum number of columns we support is {max_columns_num}."
        )



import logging
from datetime import datetime, time
from enum import Enum
from typing import Any, Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series, to_datetime

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class TimestampResolution(str, Enum):
    SECONDS = "s"
    MILLISECONDS = "ms"
    NANOSECONDS = "ns"


class UnixTimestampType(BaseType):
    weight = 4
    min_valid_year: int = 1980
    max_valid_year: int = 2100

    resolution: Optional[TimestampResolution] = None

    @staticmethod
    def _unix_timestamp_resolution(value: int) -> TimestampResolution:
        number_of_digits = len(str(int(value)))
        if number_of_digits == 10:
            return TimestampResolution.SECONDS
        if number_of_digits == 13:
            return TimestampResolution.MILLISECONDS
        if number_of_digits == 16:
            return TimestampResolution.NANOSECONDS

    @staticmethod
    def _fix_input_resolution(
        value: int, selected_resolution: TimestampResolution
    ) -> int:
        if selected_resolution == TimestampResolution.SECONDS:
            return int(value)
        if selected_resolution == TimestampResolution.MILLISECONDS:
            return int(int(value) / 1000)
        if selected_resolution == TimestampResolution.NANOSECONDS:
            return int(int(value) / 1000 / 1000)

    def _is_valid_unix_timestamp(self, value: Any) -> bool:
        if self.is_empty_value(value):
            return True

        if self.is_empty_value(value) or isinstance(value, (datetime, time)):
            return False

        if not self.resolution:
            self.resolution = self._unix_timestamp_resolution(int(value))

        if self.resolution:
            parsed_value = datetime.utcfromtimestamp(
                self._fix_input_resolution(value, self.resolution)
            )
            if self.min_valid_year <= parsed_value.year <= self.max_valid_year:
                return True
        return False

    def validate(self, data: ndarray) -> Optional[InferredField]:
        self.resolution = None
        try:
            for value in data:
                if not self._is_valid_unix_timestamp(value):
                    return None

            if not self.resolution:
                return None

            return InferredField(
                inferred_type=self.name, inferred_pattern=self.resolution
            )

        except (ValueError, OSError, OverflowError) as e:
            logger.debug(f"Cannot cast the given data to {self.name}: {e}")
            return None

    @property
    def prefer_column_casting(self) -> bool:
        return True

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        return to_datetime(column, unit=inferred_field.inferred_pattern.value)

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        if self.is_empty_value(value):
            return None
        return datetime.fromtimestamp(self._fix_input_resolution(value, pattern))

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return (np.dtype("datetime64"),)



import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from numpy import ndarray
from numpy.dtypes import (
    BoolDType,
    DateTime64DType,
    Float16DType,
    Float32DType,
    Float64DType,
    Int8DType,
    Int16DType,
    Int32DType,
    Int64DType,
    IntDType,
    ObjectDType,
    StrDType,
)
from pandas import Series

from shmessy.exceptions import FieldCastingException

from .schema import Field, InferredField
from .types.base import BaseType
from .types.boolean import BooleanType
from .types.date import DateType
from .types.datetime_ import DatetimeType
from .types.email import EmailType
from .types.float import FloatType
from .types.integer import IntegerType
from .types.ipv4_address import IPv4Type
from .types.string import StringType
from .types.unix_timestamp import UnixTimestampType

logger = logging.getLogger(__name__)


class TypesHandler:
    PACKAGE_NAME: str = "shmessy"
    TYPES_DIR: str = "types"

    def __init__(self, ignore_virtual_types: bool):
        self.__types = self._discover_types(ignore_virtual_types=ignore_virtual_types)
        self.__types_as_dict: Dict[str, BaseType] = self._types_as_dict(self.__types)

    @classmethod
    def _types_as_dict(cls, __types: List[BaseType]) -> Dict[str, BaseType]:
        res = {}
        for type_ in __types:
            res[type_.name] = type_
        return res

    @classmethod
    def _discover_types(cls, ignore_virtual_types: bool) -> List[BaseType]:
        types = [
            BooleanType(),
            DatetimeType(),
            DateType(),
            FloatType(),
            IntegerType(),
            StringType(),
            UnixTimestampType(),
        ]
        if not ignore_virtual_types:
            types += [
                IPv4Type(),
                EmailType(),
            ]
        return sorted(types, key=lambda x: x.weight)

    @staticmethod
    def _extract_bad_value(column: Series, func: Callable) -> Tuple[int, Any]:
        for idx, row in enumerate(column):
            try:
                func(row)  # noqa
            except Exception:  # noqa
                return idx + 2, row

        # If we reached this piece of code - The dtype is probably an object - do nothing!
        raise NotImplementedError()

    @staticmethod
    def _cast_with_fallback_to_null(
        value: Any, inferred_pattern: Any, type_: BaseType
    ) -> Any:
        try:
            return type_.cast_value(value, inferred_pattern)
        except Exception as e:
            logger.debug(e)
            return None

    def _fix_column(
        self,
        column: Series,
        inferred_field: InferredField,
        type_: BaseType,
        fallback_to_null: Optional[bool] = False,
    ) -> Series:
        try:
            if column.dtype.type in type_.ignore_cast_for_types():
                return column
            if fallback_to_null:
                return column.apply(
                    lambda x: self._cast_with_fallback_to_null(
                        x, inferred_field.inferred_pattern, type_
                    )
                )
            if type_.prefer_column_casting:
                return type_.cast_column(column, inferred_field)
            return column.apply(
                lambda x: type_.cast_value(x, inferred_field.inferred_pattern)
            )
        except Exception as e:
            logger.debug(f"Couldn't cast column to type {type_.name}: {e}")
            try:
                line_number, bad_value = self._extract_bad_value(
                    column=column,
                    func=lambda x: type_.cast_value(
                        value=x, pattern=inferred_field.inferred_pattern
                    ),
                )
                raise FieldCastingException(
                    type_=type_.name,
                    line_number=line_number,
                    bad_value=bad_value,
                    column_name=str(column.name),
                    pattern=inferred_field.inferred_pattern,
                )
            except NotImplementedError:
                pass

    def fix_field(
        self,
        column: Series,
        inferred_field: Field,
        fallback_to_string: bool,
        fallback_to_null: bool,
    ) -> Any:
        try:
            return self._fix_column(
                column=column,
                inferred_field=inferred_field,
                type_=self.__types_as_dict[inferred_field.inferred_type],
            )

        except FieldCastingException as e:
            logger.debug(e)

            if fallback_to_string:
                logger.debug("Could not cast the field - Apply fallback to string")
                return self._fix_column(
                    column=column, inferred_field=inferred_field, type_=StringType()
                )

            if fallback_to_null:
                return self._fix_column(
                    column=column,
                    inferred_field=inferred_field,
                    type_=self.__types_as_dict[inferred_field.inferred_type],
                    fallback_to_null=True,
                )
            raise e

    def infer_field(self, field_name: str, data: ndarray) -> Field:
        for type_ in self.__types:
            logger.debug(f"Trying to match column {field_name} to type {type_.name}")
            inferred = type_.validate(data)
            if inferred:
                return Field(
                    field_name=field_name,
                    source_type=_numpy_type_shmessy_type(data.dtype),
                    inferred_type=inferred.inferred_type,
                    inferred_pattern=inferred.inferred_pattern,
                )

        return Field(
            field_name=field_name, source_type=_numpy_type_shmessy_type(data.dtype)
        )


def _numpy_type_shmessy_type(numpy_type: Type) -> str:
    if isinstance(
        numpy_type,
        (
            IntDType,
            Int64DType,
            Int8DType,
            Int16DType,
            Int32DType,
        ),
    ):
        return IntegerType().name
    if isinstance(numpy_type, (ObjectDType, StrDType)):
        return StringType().name
    if isinstance(numpy_type, BoolDType):
        return BooleanType().name
    if isinstance(numpy_type, DateTime64DType):
        return DatetimeType().name
    if isinstance(numpy_type, (Float16DType, Float32DType, Float64DType)):
        return FloatType().name
    return str(numpy_type)



import logging
from datetime import datetime
from typing import Any, Optional, Tuple

import numpy as np
from numpy import ndarray
from pandas import Series, Timestamp, to_datetime

from ..schema import InferredField
from .base import BaseType

logger = logging.getLogger(__name__)


class DatetimeType(BaseType):
    weight = 3
    patterns: list[str] = [
        "%d-%m-%Y %H:%M",  # 11-14-2003 00:00
        "%m/%d/%y %H:%M:%S",  # 12/15/22 00:00:00
        "%m-%d-%y %H:%M:%S",  # 12-30-2022 00:00:00
        "%m/%d/%Y %H:%M:%S",  # 12/30/2022 00:00:00
        "%m-%d-%Y %H:%M:%S",  # 12-30-2022 00:00:00
        "%Y/%m/%d %H:%M:%S",  # 2022/12/30 00:00:00
        "%Y-%m-%d %H:%M:%S",  # 2022-12-30 00:00:00
        "%Y-%m-%d %H:%M:%SZ",  # 2022-12-30 00:00:00Z
        "%Y-%m-%dT%H:%M:%SZ",  # 2022-12-30T00:00:00Z
        "%Y-%m-%dT%H:%M:%S.%f",  # 2022-12-30T00:00:00.000
        "%Y-%m-%d %H:%M:%S.%fZ",  # 2022-12-30 00:00:00.000Z
        "%Y-%m-%d %H:%M:%S.%f",  # 2022-12-30 00:00:00.000
        "%Y-%m-%dT%H:%M:%S.%fZ",  # 2022-12-30T00:00:00.000Z
        "%Y-%m-%dT%H:%M:%S",  # 2022-12-30T00:00:00
        "%d/%m/%Y %H:%M",  # 30/12/2022 00:00
        "%Y-%m-%d %H:%M:%S %Z",  # 2020-09-24 11:57:27 UTC
    ]

    def validate(self, data: ndarray) -> Optional[InferredField]:
        if data.dtype.type == np.dtype("datetime64"):
            return InferredField(inferred_type=self.name)

        for pattern in self.patterns:
            valid_pattern = True
            at_least_single_not_nan_value = False
            for value in data:
                try:
                    self.cast_value(value, pattern)
                    if not self.is_empty_value(value):
                        at_least_single_not_nan_value = True
                except Exception as e:
                    logger.debug(e)
                    valid_pattern = False

            if valid_pattern and at_least_single_not_nan_value:
                return InferredField(inferred_type=self.name, inferred_pattern=pattern)

    @property
    def prefer_column_casting(self) -> bool:
        return True

    def cast_column(self, column: Series, inferred_field: InferredField) -> Series:
        return to_datetime(column, format=inferred_field.inferred_pattern)

    def cast_value(self, value: Any, pattern: Optional[Any] = None) -> Optional[Any]:
        try:
            if self.is_empty_value(value):
                return None
            if isinstance(value, (np.datetime64, Timestamp)):
                return value
            if isinstance(value, str):
                return datetime.strptime(value, pattern)
            raise Exception(f"Input value for {self.name} casting must be string.")
        except ValueError as e:
            logger.debug(f"Cannot cast the value '{value}' using pattern '{pattern}'")
            raise e

    def ignore_cast_for_types(self) -> Tuple[Any]:
        return (np.dtype("datetime64"),)



import re
from typing import BinaryIO, Dict, Optional, TextIO, Union

from pandas import DataFrame

from .schema import ShmessySchema


def _get_sampled_df(df: DataFrame, sample_size: int, random_sample: bool) -> DataFrame:
    number_of_rows: int = len(df)
    if number_of_rows < sample_size:
        sample_size = number_of_rows
    if random_sample:
        return df.sample(sample_size)
    return df.head(sample_size)


def _fix_column_names(df: DataFrame) -> Dict[str, str]:
    fixed_column_names = {}
    for column in df.columns:
        fixed_column_names[column] = re.sub("[^0-9a-zA-Z]+", "_", str(column))
    return fixed_column_names


def _fix_column_names_in_df(input_df: DataFrame, mapping: Dict[str, str]) -> DataFrame:
    return input_df.rename(columns=mapping)


def _fix_column_names_in_shmessy_schema(
    input_schema: ShmessySchema, mapping: Dict[str, str]
) -> ShmessySchema:
    for column in input_schema.columns:
        if mapping[column.field_name]:
            column.field_name = mapping[column.field_name]
    return input_schema


def _get_sample_from_csv(
    filepath_or_buffer: Union[str, TextIO, BinaryIO],
    sample_size: int,
    encoding: Optional[str] = "UTF-8",
) -> str:
    if isinstance(filepath_or_buffer, str):
        with open(file=filepath_or_buffer, mode="rt", encoding=encoding) as input_file:
            return "".join(input_file.readlines(sample_size))

    # sample = "".join(filepath_or_buffer.readlines(sample_size))
    # There is an issue with readlines(x) that reads the whole data instead of X first rows

    sample = ""
    for idx, line in enumerate(filepath_or_buffer):
        if idx > sample_size:
            break
        if not isinstance(line, str):
            line = line.decode(encoding)  # noqa
        sample += line

    filepath_or_buffer.seek(0)
    return sample
